"""
Summary catalog from fitting outputs
"""

import os
import glob

from .. import prep, utils


def set_column_formats(fit):
    """
    Set print formats for the master catalog columns
    """

    for c in ['ra', 'dec']:
        fit[c].format = '.5f'

    for col in ['MAG_AUTO', 'FLUX_RADIUS', 'A_IMAGE', 'ellipticity']:
        if col.lower() in fit.colnames:
            fit[col.lower()].format = '.2f'

    for c in ['log_risk', 'log_pdf_max', 'zq', 'chinu', 'bic_diff']:
        fit[c].format = '.2f'

    for c in ['z_risk', 'z_map', 'z02', 'z16', 'z50', 'z84', 'z97']:
        fit[c].format = '.4f'

    for c in ['t_g102', 't_g141', 't_g800l']:
        if c in fit.colnames:
            fit[c].format = '.0f'

    for c in fit.colnames:
        if c.startswith('flux_'):
            fit[c].format = '.1e'

        if c.startswith('err_'):
            fit[c].format = '.1e'

        if c.startswith('ew50_'):
            fit[c].format = '.1e'

        if c.startswith('ewhw_'):
            fit[c].format = '.1e'

        if c.startswith('sn_'):
            fit[c].format = '.1f'

        if c.startswith('zwidth'):
            fit[c].format = '.3f'

    return fit


def summary_catalog(field_root='', dzbin=0.01, use_localhost=True, filter_bandpasses=None, files=None, cdf_sigmas=None, strip_empty_columns=False, **kwargs):
    """
    Make redshift histogram and summary catalog / HTML table
    """
    import os
    import time

    import numpy as np
    from matplotlib.ticker import FixedLocator
    import matplotlib.pyplot as plt

    import astropy.table

    try:
        from .. import fitting, prep, utils
        from . import auto_script
    except:
        from grizli import prep, utils, fitting
        from grizli.pipeline import auto_script

    if filter_bandpasses is None:
        import pysynphot as S
        filter_bandpasses = [S.ObsBandpass(bpstr) for bpstr in ['acs,wfc1,f814w', 'wfc3,ir,f105w', 'wfc3,ir,f110w', 'wfc3,ir,f125w', 'wfc3,ir,f140w', 'wfc3,ir,f160w']]

    if os.path.exists('{0}.info.fits'.format(field_root)):
        orig = utils.read_catalog('{0}.info.fits'.format(field_root))
        all_files = glob.glob('{0}*full.fits'.format(field_root))
        all_files.sort()

        print('{0}.info.fits: {1} objects.  Found {2} full.fits files, checking modify dates.'.format(field_root, len(orig), len(all_files)))
        info_mtime = os.stat('{0}.info.fits'.format(field_root)).st_mtime
        keep = np.ones(len(orig), dtype=bool)

        files = []
        for file in all_files:
            id = int(file.split('_')[1].split('.full')[0])
            if id not in orig['id']:
                files.append(file)
            else:
                full_mtime = os.stat(file).st_mtime
                if full_mtime > info_mtime:
                    files.append(file)
                    keep[orig['id'] == id] = False

        orig = orig[keep]

        if len(files) == 0:
            print('Found {0}.info.fits, and {1} new objects.\n'.format(field_root, len(files)))
            return False
        else:
            print('Found {0}.info.fits.  Adding {1} new objects.\n'.format(field_root, len(files)))

    else:
        orig = None

    # SUmmary catalog
    fit = fitting.make_summary_catalog(target=field_root, sextractor=None,
                                       filter_bandpasses=filter_bandpasses,
                                       files=files, cdf_sigmas=cdf_sigmas,
                                       write_table=(dzbin is not None))
    fit.meta['root'] = field_root

    if orig is not None:
        if len(fit) > 0:
            fit = astropy.table.vstack([orig, fit])
            if dzbin is not None:
                fit.write('{0}.info.fits'.format(field_root), overwrite=True)

    mtime = []
    for i in range(len(fit)):
        full_file = '{0}_{1:05d}.full.fits'.format(fit['root'][i], fit['id'][i])
        if os.path.exists(full_file):
            mtime.append(time.ctime(os.stat(full_file).st_mtime))
        else:
            mtime.append('-')

    fit['mtime'] = mtime

    # Add photometric catalog
    try:
        catalog = glob.glob('{0}-*.cat.fits'.format(field_root))[0]
        sex = utils.GTable.gread(catalog)
        # try:
        # except:
        #     sex = utils.GTable.gread('../Prep/{0}-ir.cat.fits'.format(field_root), sextractor=True)

        idx = np.arange(len(sex))
        sex_idx = np.array([idx[sex['NUMBER'] == id][0] for id in fit['id']])

        fit['ellipticity'] = (sex['B_IMAGE']/sex['A_IMAGE'])[sex_idx]

        for col in ['MAG_AUTO', 'FLUX_RADIUS', 'A_IMAGE']:
            fit[col.lower()] = sex[col][sex_idx]
    except:
        pass

    fit = set_column_formats(fit)

    if strip_empty_columns:
        # Remove float columns with only NaN values
        #print('Strip empty columns')
        empty_cols = []
        for c in fit.colnames:
            try:
                isfin = np.isfinite(fit[c])
                if isfin.sum() == 0:
                    empty_cols.append(c)
            except:
                pass

        for c in empty_cols:
            fit.remove_column(c)

    # Just return the table if dzbin parameter not specified
    if dzbin is None:
        return fit

    # Overwrite with additional sextractor keywords
    fit.write('{0}.info.fits'.format(field_root), overwrite=True)

    clip = (fit['chinu'] < 2.0) & (fit['log_risk'] < -1)
    clip = (fit['chinu'] < 2.0) & (fit['zq'] < -3) & (fit['zwidth1']/(1+fit['z_map']) < 0.005)
    clip &= fit['bic_diff'] > 30  # -40

    bins = utils.log_zgrid(zr=[0.1, 3.5], dz=dzbin)

    fig = plt.figure(figsize=[6, 4])
    ax = fig.add_subplot(111)

    ax.hist(np.log10(1+fit['z_map']), bins=np.log10(1+bins), alpha=0.2, color='k')
    ax.hist(np.log10(1+fit['z_map'][clip]), bins=np.log10(1+bins), alpha=0.8)

    xt = np.array(np.arange(0.25, 3.55, 0.25))
    ax.xaxis.set_minor_locator(FixedLocator(np.log10(1+xt)))
    xt = np.array([1, 2, 3])
    ax.set_xticks(np.log10(1+xt))
    ax.set_xticklabels(xt)

    ax.set_xlabel('z')
    ax.set_ylabel(r'$N$')

    ax.grid()
    ax.text(0.05, 0.95, field_root, ha='left', va='top', transform=ax.transAxes)

    fig.tight_layout(pad=0.2)
    fig.savefig('{0}_zhist.png'.format(field_root))

    cols = ['root', 'mtime', 'idx', 'ra', 'dec', 'mag_auto', 't_g800l', 't_g102', 't_g141', 'z_map', 'chinu', 'bic_diff', 'zwidth1', 'd4000', 'd4000_e', 'png_stack', 'png_full', 'png_rgb', 'png_line']

    for i in range(len(cols))[::-1]:
        if cols[i] not in fit.colnames:
            cols.pop(i)

    filter_columns = ['ra', 'dec', 'mag_auto', 't_g800l', 't_g102', 't_g141', 'z_map', 'chinu', 'bic_diff', 'zwidth1', 'd4000', 'd4000_e']

    fit[cols].write_sortable_html(field_root+'-fit.html', replace_braces=True, localhost=use_localhost, max_lines=50000, table_id=None, table_class='display compact', css=None, filter_columns=filter_columns, use_json=(not use_localhost))

    fit[cols][clip].write_sortable_html(field_root+'-fit.zq.html', replace_braces=True, localhost=use_localhost, max_lines=50000, table_id=None, table_class='display compact', css=None, filter_columns=filter_columns, use_json=(not use_localhost))

    zstr = ['{0:.3f}'.format(z) for z in fit['z_map'][clip]]
    prep.table_to_regions(fit[clip], output=field_root+'-fit.zq.reg', comment=zstr)

    # if False:
    # 
    #     fit = utils.GTable.gread('{0}.info.fits'.format(root))
    #     fit = auto_script.set_column_formats(fit)
    # 
    #     cols = ['id', 'ra', 'dec', 'mag_auto', 't_g102', 't_g141', 'sn_Ha', 'sn_OIII', 'sn_Hb', 'z_map', 'log_risk', 'log_pdf_max', 'zq', 'chinu', 'bic_diff', 'zwidth1', 'png_stack', 'png_full', 'png_line']
    # 
    #     #clip = ((fit['sn_Ha'] > 5) | (fit['sn_OIII'] > 5)) & (fit['bic_diff'] > 50) & (fit['chinu'] < 2)
    #     #clip = (fit['sn_OIII'] > 5) & (fit['bic_diff'] > 100) & (fit['chinu'] < 3)
    # 
    #     test_line = {}
    #     for li in ['Ha', 'OIII', 'OII']:
    #         test_line[l] = (fit['sn_'+li] > 5) & (fit['err_'+li] < 1.e-16)
    # 
    #     clip = (test_line['Ha'] | test_line['OIII'] | test_line['OII']) & (fit['bic_diff'] > 50) & (fit['chinu'] < 2)
    # 
    #     star = fit['flux_radius'] < 2.3
    #     clip &= ~star
    # 
    #     jh = fit['mag_wfc3,ir,f125w'] - fit['mag_wfc3,ir,f160w']
    #     clip = (fit['chinu'] < 2) & (jh > 0.9) & (fit['mag_wfc3,ir,f160w'] < 23)
    #     fit['jh'] = jh
    #     fit['jh'].format = '.1f'
    # 
    #     fit['dmag'] = fit['mag_wfc3,ir,f140w'] - fit['mag_auto']
    #     fit['dmag'].format = '.1f'
    # 
    #     cols = ['idx', 'ra', 'dec', 'mag_auto', 'jh', 'dmag', 't_g141', 'sn_Ha', 'sn_OIII', 'sn_Hb', 'z_map', 'log_risk', 'log_pdf_max', 'zq', 'chinu', 'bic_diff', 'zwidth1', 'png_stack', 'png_full', 'png_line']
    # 
    #     fit[cols][clip].write_sortable_html(root+'-fit.lines.html', replace_braces=True, localhost=False, max_lines=50000, table_id=None, table_class='display compact', css=None)



"""
Helper scripts for running galfit
"""
import os
import glob

from collections import OrderedDict
import numpy as np

import astropy.io.fits as pyfits
import astropy.wcs as pywcs

from grizli import utils

try:
    from .psf import DrizzlePSF
except:
    from grizli.galfit.psf import DrizzlePSF

# Constraints
# see https://users.obs.carnegiescience.edu/peng/work/galfit/EXAMPLE.CONSTRAINTS

LINKED_CONSTRAINTS = """# small shifts, limited re, q
# linked x, y, q, pa
2  x   -4 4       # relative to initial guess
2  y   -4 4       # relative to initial guess
2  re  0.51 to 10
2  n   0.4 to 10
2  q   0.15 to 1

3  n   0.4 to 10
3  re  0.51 to 20

2  mag  12 to 35
3  mag  12 to 35

2-3 mag -3 3

3_2 x offset
3_2 y offset
3_2 q  offset
3_2 pa offset
"""

SERSIC_CONSTRAINTS = """# Component {comp}
{comp} x  -3 3
{comp} y  -3 3
{comp} mag  15 to 29
{comp} re  0.2 to 30
{comp} q   0.1 to 1.
{comp} n   0.5 to 8

"""

GALFIT_IMAGES = """===============================================================================
# IMAGE and GALFIT CONTROL PARAMETERS
A) {input}           # Input data image (FITS file)
B) {output}       # Output data image block
C) {sigma}                # Sigma image name (made from data if blank or "none")
D) {psf}   #        # Input PSF image and (optional) diffusion kernel
E) {psf_sample}                   # PSF fine sampling factor relative to data
F) {mask}                # Bad pixel mask (FITS image or ASCII coord list)
G) {constraints}                # File with parameter constraints (ASCII file)
H) 1    {xmax}   1    {ymax}   # Image region to fit (xmin xmax ymin ymax)
I) {sh}    {sh}          # Size of the convolution box (x y)
J) {ab_zeropoint:.3f}       # Magnitude photometric zeropoint
K) {ps:.3f} {ps:.3f}        # Plate scale (dx dy)    [arcsec per pixel]
O) regular             # Display type (regular, curses, both)
P) 0                   # Choose: 0=optimize, 1=model, 2=imgblock, 3=subcomps

# INITIAL FITTING PARAMETERS
#
#   For object type, the allowed functions are:
#       nuker, sersic, expdisk, devauc, king, psf, gaussian, moffat,
#       ferrer, powsersic, sky, and isophote.
#
#   Hidden parameters will only appear when they're specified:
#       C0 (diskyness/boxyness),
#       Fn (n=integer, Azimuthal Fourier Modes),
#       R0-R10 (PA rotation, for creating spiral structures).
#
# -----------------------------------------------------------------------------
#   par)    par value(s)    fit toggle(s)    # parameter description
# -----------------------------------------------------------------------------
"""


class GalfitModel(object):
    def __str__(self, wcs=None):
        s = '# Object ID: {0}\n'.format(self.id)
        s += ' 0) {0}\n'.format(self.type)
        for k in self.pidx:
            s += '{0:>2}) '.format(self.pidx[k])
            if isinstance(self.pdict[k], list):
                s += ' '.join(['{0:.4f}'.format(v) for v in self.pdict[k]])
                if k != 'output':
                    s += ' '
                    s += ' '.join(['{0:d}'.format(v) for v in self.pfree[k]])
            else:
                s += ' {0:.4f}'.format(self.pdict[k])
                if k != 'output':
                    s += ' {0:d}'.format(self.pfree[k])

            s += ' # {0}\n'.format(k)

        return s


class SersicModel(GalfitModel):
    def __init__(self, position=[0, 0], magnitude=20., r_e=1., n=4., q=0.2, pa=0, output=0, id=0, wcs=None):
        self.type = 'sersic'
        self.id = 0

        if wcs is not None:
            self.ra, self.dec = np.array(wcs.all_world2pix([position[0]],
                                                          [position[1]], 0))[0]

        else:
            self.ra = self.dec = None

        self.pidx = OrderedDict([('position', 1),
                                 ('magnitude', 3),
                                 ('r_e', 4),
                                 ('n', 5),
                                 ('q', 9),
                                 ('pa', 10),
                                 ('output', 'Z')])

        self.pdict = OrderedDict([('position', position),
                                  ('magnitude', magnitude),
                                  ('r_e', r_e),
                                  ('n', n),
                                  ('q', q),
                                  ('pa', pa),
                                  ('output', output)])

        self.pfree = OrderedDict([('position', [1, 1]),
                                 ('magnitude', 1),
                                 ('r_e', 1),
                                 ('n', 1),
                                 ('q', 1),
                                 ('pa', 1)])


class PSFModel(GalfitModel):
    def __init__(self, position=[0, 0], magnitude=20., output=0, id=0, wcs=None):
        self.type = 'psf'
        self.id = 0

        if wcs is not None:
            self.ra, self.dec = np.array(wcs.all_world2pix([position[0]],
                                                          [position[1]], 0))[0]

        else:
            self.ra = self.dec = None

        self.pidx = OrderedDict([('position', 1),
                                 ('magnitude', 3),
                                 ('output', 'Z')])

        self.pdict = OrderedDict([('position', position),
                                  ('magnitude', magnitude),
                                  ('output', output)])

        self.pfree = OrderedDict([('position', [1, 1]),
                                 ('magnitude', 1)])


class SkyModel(GalfitModel):
    def __init__(self, background=0, dx=0, dy=0, output=0, id=0, wcs=None, position=[0, 0]):
        self.type = 'sky'
        self.id = 0

        if wcs is not None:
            self.ra, self.dec = np.array(wcs.all_world2pix([position[0]],
                                                          [position[1]], 0))[0]

        else:
            self.ra = self.dec = None

        self.pidx = OrderedDict([('background', 1),
                                 ('dx', 2),
                                 ('dy', 3),
                                 ('output', 'Z')])

        self.pdict = OrderedDict([('background', background),
                                  ('dx', dx),
                                  ('dy', dy),
                                  ('output', output)])

        self.pfree = OrderedDict([('background', 1),
                                 ('dx', 1),
                                 ('dy', 1)])


SERSIC = """
# Object number: 1
 0) sersic                 #  object type
 1) {x0}  {y0}  1 1  #  position x, y
 3) 20.0890     1          #  Integrated magnitude
 4) 1.      1          #  R_e (half-light radius)   [pix]
 5) 4.      1          #  Sersic index n (de Vaucouleurs n=4)
 6) 0.0000      0          #     -----
 7) 0.0000      0          #     -----
 8) 0.0000      0          #     -----
 9) 0.7570      1          #  axis ratio (b/a)
10) -60.3690    1          #  position angle (PA) [deg: Up=0, Left=90]
 Z) 0                      #  output option (0 = resid., 1 = Don't subtract)
"""

DEVN4 = """
# Object number: 1
0) sersic                 #  object type
1) {x0}  {y0}  1 1  #  position x, y
3) 20.0890     1          #  Integrated magnitude
4) 1.      1          #  R_e (half-light radius)   [pix]
5) 4.      0          #  Sersic index n (de Vaucouleurs n=4)
6) 0.0000      0          #     -----
7) 0.0000      0          #     -----
8) 0.0000      0          #     -----
9) 0.7570      1          #  axis ratio (b/a)
10) -60.3690    1          #  position angle (PA) [deg: Up=0, Left=90]
Z) 0                      #  output option (0 = resid., 1 = Don't subtract)
"""

NUKER = """
# Object number: 1
 0) nuker                 #  object type
 1) {x0}  {y0}  1 1  #  position x, y
 3) 20.0890     1          #  Surface brightness
 4) 10.      1          #  Rb [pix]
 5) 4.2      1          #  Nuker powerlaw alpha
 6) 1.1      1          #     Nuker beta
 7) 0.5      1          #     Nuker gamma
 8) 0.0000      0          #     -----
 9) 0.7570      1          #  axis ratio (b/a)
10) -60.3690    1          #  position angle (PA) [deg: Up=0, Left=90]
 Z) 0                      #  output option (0 = resid., 1 = Don't subtract)
"""

DISK = """
# Object number: 1
 0) expdisk                 #  object type
 1) {x0}  {y0}  1 1  #  position x, y
 3) 20.0890     1          #  Surface brightness
 4) 2.      1          #  R_s [pix]
 9) 0.7570      1          #  axis ratio (b/a)
10) -60.3690    1          #  position angle (PA) [deg: Up=0, Left=90]
C0) -0.05 0                # diskyness(-)/boxyness(+)
 Z) 0                      #  output option (0 = resid., 1 = Don't subtract)
"""


def _tostr(obj):
    if isinstance(obj, list) | isinstance(obj, tuple):
        return ' '.join(['{0}'.format(o) for o in obj])
    else:
        return '{0}'.format(obj)


class GalfitObject(object):
    def __repr__(self):
        return self.__str__()

    def __getitem__(self, k):
        return self.pdict[k]

    def __setitem__(self, k, value):
        self.pdict[k] = value

    def __str__(self):
        line = '# Object\n 0) {0}\n'.format(self.name)
        for k in self.pdict:
            line += '{0:>2}) {1} \t {2} \t# {3}\n'.format(self.pidx[k], _tostr(self.pdict[k]), _tostr(self.pfree[k]), k)

        line += '{0:>2}) {1} \t# {2}\n'.format('Z', self.output, 'output option (0 = resid., 1 = Don\'t subtract)')
        return line

    def set(self, **keys):
        for k in keys:
            if k in self.pdict:
                self.pdict[k] = keys[k]

    def setfree(self, **keys):
        for k in keys:
            if k in self.pfree:
                self.pfree[k] = keys[k]


class GalfitExpdisk(GalfitObject):
    def __init__(self, output=0, fix={}, **keys):
        import copy

        self.pidx = OrderedDict([('pos', 1),
                                 ('mag', 3),
                                 ('R_s', 4),
                                 ('q', 9),
                                 ('pa', 10),
                                 ('disky', 'C0')])

        self.pdef = OrderedDict([('pos', [0, 0]),
                                    ('mag', 20.),
                                    ('R_s', 1.),
                                    ('q', 0.25),
                                    ('pa', 0),
                                    ('disky', -0.05)])

        self.pfree = OrderedDict([('pos', [1, 1]),
                                  ('mag', 1),
                                  ('R_s', 1),
                                  ('n', 1),
                                  ('q', 1),
                                  ('pa', 1),
                                  ('disky', 0)])

        self.name = 'expdisk'
        self.output = output

        self.pdict = copy.deepcopy(self.pdef)
        self.set(**keys)
        self.setfree(**fix)


class GalfitSersic(GalfitObject):
    def __init__(self, output=0, disk=False, dev=False, gaussian=False, fix={}, **keys):
        """
        Sersic profile.

        Options to fix n for:
                disk, n=1
                 dev, n=4
            gaussian, n=0.5

        """
        import copy

        self.pidx = OrderedDict([('pos', 1),
                                 ('mag', 3),
                                 ('R_e', 4),
                                 ('n', 5),
                                 ('q', 9),
                                 ('pa', 10)])

        self.pdef = OrderedDict([('pos', [0, 0]),
                                    ('mag', 20.),
                                    ('R_e', 1.),
                                    ('n', 4),
                                    ('q', 0.25),
                                    ('pa', 0)])

        self.pfree = OrderedDict([('pos', [1, 1]),
                                  ('mag', 1),
                                  ('R_e', 1),
                                  ('n', 1),
                                  ('q', 1),
                                  ('pa', 1)])

        self.name = 'sersic'
        self.output = output

        self.pdict = copy.deepcopy(self.pdef)
        self.set(**keys)
        self.setfree(**fix)

        if disk:
            self.set(n=1)
            self.setfree(n=0)
        elif dev:
            self.set(n=4)
            self.setfree(n=0)
        elif gaussian:
            self.set(n=0.5)
            self.setfree(n=0)


class GalfitKing(GalfitObject):
    """King profile"""

    def __init__(self, output=0, fix={}, **keys):
        import copy

        self.pidx = OrderedDict([('pos', 1),
                                 ('mu', 3),
                                 ('Rc', 4),
                                 ('Rt', 5),
                                 ('alpha', 6),
                                 ('q', 9),
                                 ('pa', 10)])

        self.pdef = OrderedDict([('pos', [0, 0]),
                                    ('mu', 20.),
                                    ('Rc', 1.),
                                    ('Rt', 4),
                                    ('alpha', 2.0),
                                    ('q', 0.9),
                                    ('pa', 0)])

        self.pfree = OrderedDict([('pos', [1, 1]),
                                  ('mu', 1),
                                  ('Rc', 1),
                                  ('Rt', 1),
                                  ('alpha', 1),
                                  ('q', 1),
                                  ('pa', 1)])

        self.name = 'king'
        self.output = output

        self.pdict = copy.deepcopy(self.pdef)
        self.set(**keys)
        self.setfree(**fix)


class GalfitPSF(GalfitObject):
    def __init__(self, output=0, fix={}, **keys):
        import copy

        self.pidx = OrderedDict([('pos', 1),
                                 ('mag', 3)])

        self.pdef = OrderedDict([('pos', [0, 0]),
                                  ('mag', 20.)])

        self.pfree = OrderedDict([('pos', [1, 1]),
                                  ('mag', 1)])

        self.name = 'psf'
        self.output = output

        self.pdict = copy.deepcopy(self.pdef)
        self.set(**keys)
        self.setfree(**fix)


class GalfitSky(GalfitObject):
    def __init__(self, output=0, fix={}, **keys):
        import copy

        self.pidx = OrderedDict([('bg', 1),
                                 ('dx', 2),
                                 ('dy', 3)])

        self.pdef = OrderedDict([('bg', 0.),
                                  ('dx', 0.),
                                  ('dy', 0.)])

        self.pfree = OrderedDict([('bg', 1),
                                  ('dx', 0),
                                  ('dy', 0)])

        self.name = 'sky'
        self.output = output

        self.pdict = copy.deepcopy(self.pdef)
        self.set(**keys)
        self.setfree(**fix)


class Galfitter(object):
    def __init__(self, root='sdssj1723+3411', filter='f140w', galfit_exec='galfit', catfile=None, segfile=None):
        self.root = root
        self.galfit_exec = galfit_exec

        self.filter = filter
        data = self.get_data(filter=filter, catfile=catfile, segfile=segfile)
        self.sci, self.wht, self.seg, self.cat, self.wcs = data

        self.flt_files = self.get_flt_files(sci_hdu=self.sci)
        self.DPSF = DrizzlePSF(flt_files=self.flt_files, info=None, driz_image=self.sci.filename())

    def get_data(self, filter='f140w', catfile=None, segfile=None):
        import glob

        sci_file = glob.glob('{0}-{1}_dr?_sci.fits'.format(self.root, filter))

        sci = pyfits.open(sci_file[0])
        wht = pyfits.open(sci_file[0].replace('sci', 'wht'))

        if segfile is None:
            segfile = sci_file[0].split('_dr')[0]+'_seg.fits'

        seg = pyfits.open(segfile)

        if catfile is None:
            catfile = '{0}-{1}.cat.fits'.format(self.root, filter)

        cat = utils.GTable.gread(catfile)

        wcs = pywcs.WCS(sci[0].header, relax=True)
        return sci, wht, seg, cat, wcs

    def get_flt_files(self, sci_hdu=None):
        flt_files = []
        h = self.sci[0].header
        for i in range(h['NDRIZIM']):
            file = h['D{0:03d}DATA'.format(i+1)].split('[sci')[0]
            if file not in flt_files:
                flt_files.append(file)

        return flt_files

    @staticmethod
    def fit_arrays(sci, wht, seg, psf, id=None, platescale=0.06, exptime=0, path='/tmp', root='gf', galfit_exec='galfit', use_subprocess=False, ab_zeropoint=26, gaussian_guess=False, components=[GalfitSersic()], recenter=True, psf_sample=1, write_constraints=True):
        """
        Fit array data with Galfit

        """
        import subprocess
        from astropy.coordinates import Angle
        import astropy.units as u

        rms = 1/np.sqrt(wht)  # *exptime
        if exptime > 0:
            rms = np.sqrt((rms*exptime)**2+sci*exptime*(sci > 0))/exptime

        rms[wht == 0] = 1e30

        if id is not None:
            mask = ((seg > 0) & (seg != id)) | (wht == 0)
        else:
            mask = wht == 0

        sh = sci.shape[0]

        fp = open('{0}/{1}.feedme'.format(path, root), 'w')
        sci_file = '{0}/{1}_sci.fits'.format(path, root)
        constraints_file = sci_file.replace('_sci.fits', '.constraints')

        cmd = GALFIT_IMAGES.format(input=sci_file,
                                   output=sci_file.replace('_sci', '_out'),
                                   sigma=sci_file.replace('_sci', '_rms'),
                                   psf=sci_file.replace('_sci', '_psf'),
                                   mask=sci_file.replace('_sci', '_mask'),
                                   constraints=constraints_file,
                                   xmax=sh, ymax=sh, sh=sh,
                                   ps=platescale, psf_sample=psf_sample,
                                   ab_zeropoint=ab_zeropoint)

        fp.write(cmd)

        if gaussian_guess:
            fit, q, theta = fit_gauss(sci)

        if (not os.path.exists(constraints_file)) & (write_constraints):
            fpc = open(constraints_file, 'w')
        else:
            fpc = None

        for i, comp in enumerate(components):
            if recenter:
                comp.set(pos=[sh/2., sh/2.])

            if gaussian_guess:
                comp.set(q=q, pa=Angle.wrap_at(theta*u.rad, 360*u.deg).to(u.deg).value)

            fp.write(str(comp))

            if fpc is not None:
                fpc.write(SERSIC_CONSTRAINTS.format(comp=i+1))

        fp.close()
        if fpc is not None:
            fpc.close()

        pyfits.writeto(sci_file, data=sci, overwrite=True)
        pyfits.writeto(sci_file.replace('_sci', '_rms'),
                       data=rms, overwrite=True)
        pyfits.writeto(sci_file.replace('_sci', '_mask'),
                       data=mask*1, overwrite=True)
        pyfits.writeto(sci_file.replace('_sci', '_psf'),
                       data=psf, overwrite=True)

        for ext in ['out', 'model']:
            fi = sci_file.replace('_sci', '_'+ext)
            if os.path.exists(fi):
                os.remove(fi)

        call_string = '{0} {1}/{2}.feedme'.format(galfit_exec, path, root)
        if use_subprocess:
            p = subprocess.Popen(call_string.split(), stdout=subprocess.PIPE)
            stdout_output = p.communicate()[0]
            with open('{0}/{1}.stdout.txt'.format(path, root), 'w') as fp:
                fp.write(stdout_output.decode('utf-8'))
                
        else:
            os.system(call_string)

        # Move galfit.[latest] to output file
        gf_files = glob.glob('galfit.[0-9]*')
        gf_files.sort()
        gf_log = sci_file.replace('_sci.fits', '.gfmodel')

        # Results dictionary
        res = {}
        res['rms'] = rms
        res['mask'] = mask*1
        im = pyfits.open(sci_file.replace('_sci', '_out'))
        res['model'] = im[2]
        res['resid'] = im[3]
        
        if len(gf_files) > 0:
            os.system('mv {0} {1}'.format(gf_files[-1], gf_log))
            lines = open(gf_log).readlines()
            res['ascii'] = lines

        return res

    def fit_object(self, id=449, radec=(None, None), size=40, components=[GalfitSersic()], recenter=True, get_mosaic=True, gaussian_guess=False, get_extended=True):
        """
        Fit an object
        """
        if id is not None:
            rd = self.cat['X_WORLD', 'Y_WORLD'][self.cat['NUMBER'] == id][0]
            radec = tuple(rd)

        xy = self.wcs.all_world2pix([radec[0]], [radec[1]], 0)
        xy = np.array(xy).flatten()
        xp = np.asarray(np.round(xy),dtype=int)

        slx, sly, wcs_slice = self.DPSF.get_driz_cutout(ra=radec[0], dec=radec[1], get_cutout=False, size=size)
        #drz_cutout = self.get_driz_cutout(ra=ra, dec=dec, get_cutout=True)
        h = self.sci[0].header
        psf = self.DPSF.get_psf(ra=radec[0], dec=radec[1], filter=self.filter.upper(), wcs_slice=wcs_slice, pixfrac=h['D001PIXF'], kernel=h['D001KERN'], get_extended=get_extended)

        exptime = h['EXPTIME']
        sci = self.sci[0].data[sly, slx]  # *exptime
        wht = self.wht[0].data[sly, slx]
        rms = 1/np.sqrt(wht)  # *exptime
        rms = np.sqrt((rms*exptime)**2+sci*exptime*(sci > 0))/exptime

        rms[wht == 0] = 1e30

        seg = self.seg[0].data[sly, slx]
        if id is not None:
            mask = ((seg > 0) & (seg != id)) | (wht == 0)
        else:
            mask = wht == 0

        sh = sci.shape[0]
        path = '/tmp/'

        psf_file = '{0}-{1}_psf_{2:05d}.fits'.format(self.root, self.filter, id)

        fp = open(path+'galfit.feedme', 'w')

        fp.write(GALFIT_IMAGES.format(input=path+'gf_sci.fits', output=path+'gf_out.fits', sigma=path+'gf_rms.fits', psf=psf_file, mask=path+'gf_mask.fits', xmax=sh, ymax=sh, sh=sh, ps=self.DPSF.driz_pscale, psf_sample=1, constraints='none', ab_zeropoint=26))

        if gaussian_guess:
            fit, q, theta = fit_gauss(sci)

        from astropy.coordinates import Angle
        import astropy.units as u

        for comp in components:
            if recenter:
                comp.set(pos=[sh/2., sh/2.])

            if gaussian_guess:
                comp.set(q=q, pa=Angle.wrap_at(theta*u.rad, 360*u.deg).to(u.deg).value)

            fp.write(str(comp))

        fp.close()

        pyfits.writeto(path+'gf_sci.fits', data=sci, overwrite=True)
        pyfits.writeto(path+'gf_rms.fits', data=rms, overwrite=True)
        pyfits.writeto(path+'gf_mask.fits', data=mask*1, overwrite=True)
        #pyfits.writeto(path+'gf_psf.fits', data=psf[1].data, overwrite=True)

        pyfits.writeto(psf_file, data=psf[1].data, overwrite=True)

        for ext in ['out', 'model']:
            if os.path.exists(path+'gf_{0}.fits'.format(ext)):
                os.remove(path+'gf_{0}.fits'.format(ext))

        os.system('{0} {1}/galfit.feedme'.format(self.galfit_exec, path))

        #out = pyfits.open('/tmp/gf_out.fits')

        # Get in DRZ frame
        gf_file = glob.glob('galfit.[0-9]*')[-1]
        lines = open(gf_file).readlines()
        for il, line in enumerate(lines):
            if line.startswith('A)'):
                lines[il] = 'A) {0}     # Input data image (FITS file)\n'.format(self.sci.filename())
            if line.startswith('B)'):
                lines[il] = 'B) {0}-{1}_galfit_{2:05d}.fits      # Output data image block\n'.format(self.root, self.filter, id)
            if line.startswith('P) 0'):
                lines[il] = "P) 1                   # Choose: 0=optimize, 1=model, 2=imgblock, 3=subcomps\n"

            if line.startswith('H)'):
                out_sh = self.sci[0].data.shape
                lines[il] = "H) 1    {0}   1    {1}   # Image region to fit (xmin xmax ymin ymax)\n".format(out_sh[1], out_sh[0])

            if line.startswith(' 1)'):
                xy = np.asarray(line.split()[1:3],dtype=float)
                lines[il] = ' 1) {0}  {1}  1 1  #  Position x, y\n'.format(xy[0]+slx.start, xy[1]+sly.start)

        fp = open('galfit.{0}.{1:05d}'.format(self.filter, id), 'w')
        fp.writelines(lines)
        fp.close()

        os.system('{0} galfit.{1}.{2:05d}'.format(self.galfit_exec, self.filter, id))

        model = pyfits.open('{0}-{1}_galfit_{2:05d}.fits'.format(self.root, self.filter, id), mode='update')
        model[0].data /= self.sci[0].header['EXPTIME']
        model.flush()

        return(lines, model)

        model = pyfits.open('/tmp/gf_model.fits')
        if False:
            full_model = self.sci[0].data*0

        full_model += model[0].data
        pyfits.writeto('{0}-{1}_galfit.fits'.format(self.root, self.filter), data=full_model, header=self.sci[0].header, overwrite=True)


def fit_gauss(sci):
    from astropy.modeling import models, fitting

    sh = sci.shape
    gau = models.Gaussian2D(amplitude=sci.max(), x_mean=sh[0]/2, y_mean=sh[0]/2., x_stddev=None, y_stddev=None, theta=None, cov_matrix=None)

    lm = fitting.LevMarLSQFitter()

    yp, xp = np.indices(sh)
    fit = lm(gau, xp, yp, sci)

    q = (fit.x_stddev/fit.y_stddev)**2
    theta = fit.theta.value
    if q > 1:
        q = 1./q
        theta += np.pi/2.

    return fit, q, theta



import os
import inspect
import warnings

import scipy.ndimage as nd
import numpy as np

import astropy.io.fits as pyfits
import astropy.time
import astropy.units as u

from jwst import datamodels
from jwst.ramp_fitting.ramp_fit_step import get_reference_file_subarrays
from jwst import __version__ as CAL_VER

try:
    import sep
    HAS_SEP = True
except ImportError:
    HAS_SEP = False
    
import mastquery.utils
from grizli import utils

DET1PIPE = None

DQ_ANY = 1
DQ_NAN = 8
DQ_NJUMP = 16
DQ_HOT = 32
DQ_RESID = 64
DQ_ERR = 128
DQ_FIRST_SATURATED = 256

BKG_KWARGS = dict(bw=128, bh=128, fw=3, fh=3)

SCALE_RNOISE = {'xxxNRSIRS2RAPID': 2.}

PIXELDQ_BITS = {'NRS1': [10], 
                'NRS2': [10],
                'NIS': [10,11,13,14],
                }
                
for d in [1,2,3,4,'LONG']:
    for m in 'AB':
        PIXELDQ_BITS[f'NRC{m}{d}'] = [10,11,13,14]

HOT_THRESHOLDS = {'NRS1': [1000,20000],
                  'NRS2': [1000,20000],
                   'NIS': [40000,40000]
                }
                
for d in ['LONG']:
    for m in 'AB':
        HOT_THRESHOLDS[f'NRC{m}{d}'] = [40000,40000]

for d in [1,2,3,4]:
    for m in 'AB':
        HOT_THRESHOLDS[f'NRC{m}{d}'] = [1000,20000]

# Global CRDS_CONTEXT

if 'CRDS_CONTEXT' in os.environ:
    CRDS_CONTEXT = os.getenv('CRDS_CONTEXT')
else:
    CRDS_CONTEXT = 'jwst_1069.pmap'
    # CRDS_CONTEXT = 'jwst_1123.pmap'


def get_pipeline_object():
    """
    Initialize `jwst.pipeline.Detector1Pipeline`
    """
    from jwst.pipeline import Detector1Pipeline

    global DET1PIPE
    
    if DET1PIPE is None:
        DET1PIPE = Detector1Pipeline()
    
    return DET1PIPE


def download_files(file):
    """
    Download uncal and rate files
    """

    if not os.path.exists(file):
        _ = mastquery.utils.download_from_mast([file])
    if not os.path.exists(file.replace('_uncal','_rate')):
        _ = mastquery.utils.download_from_mast([file.replace('_uncal','_rate')])
    

def process_level1_to_dark(file):
    """
    Run `jwst.pipeline.Detector1Pipeline` up through `dark_current`
    
    https://jwst-pipeline.readthedocs.io/en/latest/jwst/pipeline/calwebb_detector1.html
    
    """
    
    pipe = get_pipeline_object()
    
    download_files(file)
    
    if CRDS_CONTEXT:
        os.environ["CRDS_CONTEXT"] = CRDS_CONTEXT
        
    grp = pipe.group_scale.process(file)
    dqi = pipe.dq_init.process(grp)
    sat = pipe.saturation.process(dqi)
    bias = pipe.superbias.process(sat)
    refpix = pipe.refpix.process(bias)
    lin = pipe.linearity.process(refpix)
    pers = pipe.persistence.process(lin)
    dark = pipe.dark_current.process(pers)
    
    return dark


def process_uncal_level1(file='jw01208048001_03101_00001_nrs1_uncal.fits', output_extension='_xrate', jump_threshold=4, jump_ndilate=1, erode_snowballs=5, grow_snowballs=5, resid_thresh=4, hot_thresh='auto', hot_type='diff', max_njump=6, groups_for_rnoise=np.inf, flag_for_persistence=True, outlier_min_nints=3, integration_sigmas=[5,4,3], rescale_uncertainty=True, rescale_with_background=True, bkg_kwargs=BKG_KWARGS, dark=None, verbose=True, debug=None, **kwargs):
    """
    Custom ramp-fit scripts for JWST uncal images.

    All calculations are done with `numpy` arrays to be somewhat faster than 
    computing least-squares fits to individual pixel ramps.  One shortcut here
    is that only ramp sequences with two or fewer identified jumps are 
    considered for the ramp fit, i.e., the reads before the first jump and 
    after and including the last jump.
    
    Parameters
    ----------
    file : str
        JWST raw "uncal" filename
    
    output_extension : str
        The script puts the derived ramp and uncertainty into a ``rate``-like 
        file with filename ``file.replace('_uncal', output_extension)``.
        
    jump_threshold : float
        Jump identification threshold, sigma
    
    jump_ndilate : int
        Number of `scipy.ndimage.binary_dilation` iterations run on the initial
        jump mask
    
    erode_snowballs : int
        Number of `scipy.ndimage.binary_erosion` iterations run on the jump
        mask of each read before identifying snowballs.  This removes small
        clumps of pixels that probably aren't big snowballs
    
    grow_snowballs : float
        Once snowballs are identified as big clumps of jump pixels, compute
        the pixel area of each smowball clump and make a circular mask with
        radius ``grow_snowballs * np.sqrt(area/pi)``
    
    resid_thresh : float
        Compute a ramp model and flag pixels where
        ``(ramp - model) > resid_thresh*sigma``
    
    hot_thresh : [float, float]
        Thresholds for hot pixels.  The first number is the DN threshold
        in the first ramp read.  The second number is the threshold in the 
        first ramp *diff*.
    
    hot_type : ('diff', 'raw')
        If ``diff``, then first hot pixel threshold is on the first ramp 
        *difference*, otherwise threshold is on the raw values of the first
        read
    
    max_njump : int
        Unset jump mask where more than ``max_jump`` jumps are found in a ramp
    
    groups_for_rnoise : int
        Minimum `ngroups` for estimating read noise directy from the ramp fits
    
    flag_for_persistence : bool
        While stepping through integrations and ramps, make a mask for flagged
        saturated pixels and set the flag for *all* subsequent reads of that
        pixel, i.e., to catch saturated pixels that will likely cause 
        persistence in subsequent integrations.
    
    outlier_min_nints : int
        Run outlier rejection across integrations when at least
        ``outlier_min_nints`` integrations are available.
    
    integration_sigmas : list of float
        Iterative thresholds of integration outlier rejection
    
    rescale_uncertainty : bool
        Rescale final uncertainties based on statistics of the final rate
        image
    
    dark : `jwst.datamodels.ramp.RampModel`
        Precomputed ramp product.  If not provided, run `calwebb_detector1 <https://jwst-pipeline.readthedocs.io/en/latest/jwst/pipeline/calwebb_detector1.html>`_ 
        up to the `dark_current <https://jwst-pipeline.readthedocs.io/en/latest/jwst/dark_current/index.html#dark-current-step>`_ step.
    
    verbose : bool
        Message verbosity
    
    Returns
    -------
    debug : dict
        Dictionary with a bunch of stuff computed by the script
        
    """
    
    frame = inspect.currentframe()
    _LOGFILE = utils.LOGFILE
    warnings.filterwarnings('ignore', category=RuntimeWarning)
    
    if 'grow_snowball' in kwargs:
        grow_snowballs = kwargs['grow_snowball']
        
    utils.LOGFILE = file.split('.fits')[0] + '.log'
    utils.log_function_arguments(utils.LOGFILE, frame, 'process_uncal',
                                 ignore=['dark','debug','_LOGFILE',
                                         'CRDS_CONTEXT'],
                                 )
    
    pipe = get_pipeline_object()
    
    if debug is not None:
        if debug['file'] == file:
            print('debug: use dark from debug data')
            dark = debug['dark']
            
    if dark is None:
        # Detector1 through dark_current
        # https://jwst-pipeline.readthedocs.io/en/latest/jwst/pipeline/calwebb_detector1.html
        # grp = pipe.group_scale.process(file)
        # dqi = pipe.dq_init.process(grp)
        # sat = pipe.saturation.process(dqi)
        # bias = pipe.superbias.process(sat)
        # refpix = pipe.refpix.process(bias)
        # lin = pipe.linearity.process(refpix)
        # pers = pipe.persistence.process(lin)
        # dark = pipe.dark_current.process(pers)
        
        dark = process_level1_to_dark(file)
    
    NINTS, NGROUPS, NYPIX, NXPIX = dark.data.shape
    
    ## readnoise, gain reference files
    readnoise_filename = pipe.get_reference_file(dark, 'readnoise')
    gain_filename = pipe.get_reference_file(dark, 'gain')
    # mask_filename = pipe.get_reference_file(dark, 'mask')
    
    msg =  f'\nprocess_uncal: {file}   readnoise file {readnoise_filename}'
    msg += f'\nprocess_uncal: {file}   gain file      {gain_filename}'
    utils.log_comment(utils.LOGFILE, msg, show_date=False, verbose=verbose)
    
    ## From jwst.ramp_fitting.ramp_fit_step
    with datamodels.ReadnoiseModel(readnoise_filename) as readnoise_model, \
            datamodels.GainModel(gain_filename) as gain_model:

        # Try to retrieve the gain factor from the gain reference file.
        # If found, store it in the science model meta data, so that it's
        # available later in the gain_scale step, which avoids having to
        # load the gain ref file again in that step.
        if gain_model.meta.exposure.gain_factor is not None:
            gain_factor = gain_model.meta.exposure.gain_factor
            dark.meta.exposure.gain_factor = gain_factor

        # Get gain arrays, subarrays if desired.
        frames_per_group = dark.meta.exposure.nframes
        readnoise_2d, gain_2d = get_reference_file_subarrays(
            dark, readnoise_model, gain_model, frames_per_group)
            
    ### Now comes the good stuff
    
    if dark.meta.exposure.readpatt in SCALE_RNOISE:
        msg = f'process_uncal: {file} scale readnoise_2d for '
        msg += f'READPATT = {dark.meta.exposure.readpatt}: '
        msg += f'{SCALE_RNOISE[dark.meta.exposure.readpatt]:.2f}'
        
        utils.log_comment(utils.LOGFILE, msg, show_date=False, verbose=verbose)
        readnoise_2d *= SCALE_RNOISE[dark.meta.exposure.readpatt]
        
    ## Read times, correct??
     
    int1 = dark.group['integration_number'] == 1
    grp = dark.group[int1]
    
    t = astropy.time.Time(grp['helio_end_time'], format='mjd')
    
    try:
        #if len(t) == len(grp):
        #     print('\n debug !! len(t) == len(dark.group) !! (NIRSPEC?)')
        #     dt = t[1:] - (t[0] - dark.meta.exposure.frame_time*u.second)
        #     times = [dark.meta.exposure.frame_time]
        #     for i in range(dark.meta.exposure.ngroups-1):
        #         times.append(dt[i].to(u.second).value)
        # else:
        #     dt = t[1:] - (t[0] - dark.meta.exposure.frame_time*u.second)
        #     times = []
        #     for i in range(dark.meta.exposure.ngroups):
        #         times.append(dt[i].to(u.second).value)
        dt = t[1:] - (t[0]) # - dark.meta.exposure.frame_time*u.second)
        times = []
        for i in range(dark.meta.exposure.ngroups-1):
            times.append(dt[i].to(u.second).value)
        
        times.append(times[-1] + np.diff(times)[-1])
        
    except IndexError:
        #print('xxx', len(t), dark.meta.exposure.ngroups)
        
        print('!!! Times failed')
        return {'file':file, 'dark':dark}
        
    times = np.array(times)
    
    msg = f'process_uncal: {file}  group_time = {dark.meta.exposure.group_time:.1f}'
    msg += f'\nprocess_uncal: {file}  frame_time = {dark.meta.exposure.frame_time:.1f}'
    for i in range(NGROUPS):
        msg += f'\nprocess_uncal: {file}  sample[{i:>2}] = {times[i]:>6.1f}'

    msg += f'\nprocess_uncal: {file}  integration_time = {dark.meta.exposure.integration_time:6.1f}'
        
    utils.log_comment(utils.LOGFILE, msg, show_date=False, verbose=verbose)
    
    # print('xxx', times, np.diff(times))
    
    # Copy dark dq array
    groupdq = dark.groupdq*1
        
    jump_saturated = np.zeros((NYPIX, NXPIX), dtype=groupdq.dtype)
    
    # Containers across integrations

    sci_list = []
    err_list = []
    dq_list = []
    slope_list = []
    slope_err_list = []
    nsamp_list = []
    rn_list = []
    jump_list = []
    
    for integ in range(NINTS):
        msg = f'\nprocess_uncal: {file} run for int {integ+1} / {NINTS}'
        utils.log_comment(utils.LOGFILE, msg, show_date=False, verbose=verbose)
        
        ## Array copies
        dd = dark.data[integ,:,:,:] #/gain_2d
        #dd[groupdq[integ,:,:,:] > 0] = np.nan
        
        ## Diffs for jump indentification
        ds = np.diff(dd, axis=0)
        #ds[groupdq[integ,1:,:,:] > 0] = np.nan
        
        ## times and index arrays in full cube like the ramp
        t2d = np.zeros_like(dd)
        i2d = np.zeros((NGROUPS, NYPIX, NXPIX), dtype=int)
        for i, t in enumerate(times):
            t2d[i,:,:] = t
            i2d[i,:,:] = i
        
        ## global "background" from the first read
        bkg0 = np.nanmedian(dark.data[integ,0,:,:])

        ## Crude uncertainty arrays
        rn_det = readnoise_2d / np.sqrt(dark.meta.exposure.nframes)
        raw_var = ((rn_det)**2 + np.maximum(dd - bkg0, 0)/gain_2d)
        raw_err = np.sqrt(raw_var[:-1,:,:] + raw_var[1:,:,:])
        
        ## Identify jumps
        msg = f'process_uncal: {file} {integ+1} find jumps'
        utils.log_comment(utils.LOGFILE, msg, show_date=False, verbose=verbose)
        
        jump = ds > (np.nanmedian(ds, axis=0)+jump_threshold*raw_err)
        njump = jump.sum(axis=0)
        
        ## Jumps only if first read not flagged
        jump[0,:,:] &= (groupdq[integ,0,:,:] == 0)
    
        ## Dilate jump
        msg = f'process_uncal: {file} {integ+1} dilate jumps'
        utils.log_comment(utils.LOGFILE, msg, show_date=False, verbose=verbose)
        
        for i in range(NGROUPS-1):
            ji = jump[i,:,:] & (njump < max_njump)
            jump[i,:,:] |= nd.binary_dilation(ji,
                                              structure=np.ones((3,3)),
                                              iterations=jump_ndilate)

        ## Identify snowballs and grow masks around them
        for ij in range(NGROUPS-1):
            ji = jump[ij,:,:]

            js = nd.binary_erosion(ji, iterations=erode_snowballs)

            label, num_labels = nd.label(js)
            if num_labels == 0:
                ## No snowballs identified
                msg = f'process_uncal: {integ+1} snowball read {ij+1:>2} '
                msg += f' n={num_labels:>2}  '
                msg += f'nsat={int(jump_saturated.sum())//4:>6}'
                utils.log_comment(utils.LOGFILE, msg, show_date=False, 
                                  verbose=verbose)
                
                continue
        
            xy_label = nd.center_of_mass(label, labels=label, 
                                         index=np.unique(label)[1:])
            area = nd.sum(label > 0, labels=label, index=np.unique(label)[1:])

            R_label = np.sqrt(area/np.pi)
            yp, xp = np.indices((NYPIX, NXPIX))

            lmask = np.zeros_like(js)
            smask = np.zeros_like(js)

            for (yyi, xxi), ri in zip(xy_label, R_label):
                R = np.sqrt((xp-xxi)**2 + (yp-yyi)**2)
                lmask |= R < grow_snowballs*np.maximum(ri, 2)
                smask |= R < grow_snowballs/2*np.maximum(ri, 2)
    
            jump[ij,:,:] |= lmask
            
            ## Mask for saturated pixels
            blob_sat = (groupdq[integ,ij,:,:] > 0) | smask
            # blob_sat = nd.binary_erosion(blob_sat, iterations=1)
            blob_sat = nd.binary_dilation(blob_sat, iterations=2)*2
            #blob_sat = smask*2
            
            jump_saturated |= blob_sat.astype(jump_saturated.dtype)
            
            msg = f'process_uncal: {integ+1} snowball read {ij+1:>2} '
            msg += f' n={num_labels:>2}  '
            msg += f'nsat={int(jump_saturated.sum())//4:>6}'
            utils.log_comment(utils.LOGFILE, msg, show_date=False, 
                              verbose=verbose)

            if flag_for_persistence:
                groupdq[integ,ij,:,:] |= jump_saturated
        
        ## Number of jumps along the ramp
        njump = jump.sum(axis=0)
            
        ## Fit slopes when just zero or one jump found
        
        # Array index of first jump
        t2x = t2d[1:,:,:]*0+times[-1]*2
        t2x[jump > 0] = t2d[1:,:,:][jump > 0]
        ix = np.nanargmin(t2x, axis=0)
        
        # No jumps found
        ix[njump == 0] = NGROUPS
        
        # Array index after jump
        t2x = t2d[1:,:,:]*0
        t2x[jump > 0] = t2d[1:,:,:][jump > 0]
        ix2 = np.nanargmax(t2x, axis=0)
    
        n2d = np.zeros_like(i2d)
        ix2d = np.zeros_like(i2d)
        ix2d2 = np.zeros_like(i2d)
        for i, t in enumerate(times):
            n2d[i,:,:] = njump
            ix2d[i,:,:] = ix
            ix2d2[i,:,:] = ix2
    
        dvalid = np.zeros((NGROUPS, NYPIX, NXPIX), dtype=bool)
        
        # Masks for before / after jumps
        before = (i2d <= ix) & (ix2d > 1) & (groupdq[integ,:,:,:] == 0)
        after = (i2d > ix2) & (ix2d2 < NGROUPS-2) #& (n2d == 1)
        after &= (groupdq[integ,:,:,:] == 0)

        tmx = np.zeros_like(dd) #*np.nan
        dmx = np.zeros_like(dd) #*np.nan
        xme = np.zeros_like(dd) #*np.nan
        yme = np.zeros_like(dd) #*np.nan

        for msk, mask_type in zip([before, after],
                                  ['before first jump', 'after last jump']):
            
            msg = f'process_uncal: {file} {integ+1} slope mask {mask_type}'
            utils.log_comment(utils.LOGFILE, msg, show_date=False, 
                              verbose=verbose)
        
            ddm = dd*1
            ddm[~msk] = np.nan
            tm = t2d*1
            tm[~msk] = np.nan
            ymean = np.nanmean(ddm, axis=0)
            xmean = np.nanmean(tm, axis=0)
    
            dvalid[msk] = True
    
            tmx[msk] = (tm-xmean)[msk]
            dmx[msk] = (ddm-ymean)[msk]
    
            xme[msk] = (tm*0 + xmean)[msk]
            yme[msk] = (ddm*0 + ymean)[msk]
        
        msg = f'process_uncal: {file} {integ+1} OLS slope model'
        utils.log_comment(utils.LOGFILE, msg, show_date=False, verbose=verbose)

        ## Simple ordinary least squares fit
        slope_num = np.nansum(tmx*dmx, axis=0)
        slope_den = np.nansum(tmx**2, axis=0)
        slope = slope_num/slope_den
        #slope_err = 1./np.sqrt(slope_den)
        
        ## Intercept
        ahat = yme - slope*xme
        # Model ramp
        smodel = ahat + slope*t2d
        
        ## Recompute with weights after removing the intercept
        ddm = (dd - ahat)
        ddm[~dvalid] = np.nan
        tm = t2d*1
        tm[~dvalid] = np.nan
        
        slope_num = np.nansum(ddm*tm/raw_var, axis=0)
        slope_den = np.nansum(tm**2/raw_var, axis=0)
        slope = slope_num/slope_den
        slope_err = 1/np.sqrt(slope_den)
        
        ## Output arrays
        sci = slope*1
        dq = (~np.isfinite(sci))*DQ_NAN | (njump > 4)*DQ_NJUMP
        dq |= (groupdq[integ,0,:,:] > 0)*DQ_FIRST_SATURATED
    
        sci[njump > 4] = np.nan
        sci[~np.isfinite(sci)] = 0.

        nsamp = dvalid.sum(axis=0)
        
        ## Residuals w.r.t model
        resid = (dd - smodel)
    
        xerr = np.sqrt((readnoise_2d)**2 + slope*t2d/gain_2d)
        nbad = ((np.abs(resid) > resid_thresh*xerr) & (xme > 0)).sum(axis=0)
        
        msg = f'process_uncal: {file} {integ+1} residuals > {resid_thresh}: '
        msg += f' {(nbad > 1).sum()}'
        utils.log_comment(utils.LOGFILE, msg, show_date=False, verbose=verbose)
        
        if hot_thresh in ['auto']:
            if dark.meta.instrument.detector in HOT_THRESHOLDS:
                hot_thresh = HOT_THRESHOLDS[dark.meta.instrument.detector]
                
        ## Hot pixels
        if hot_type == 'raw':
            hot = (dark.data[integ,0,:,:] > hot_thresh[0]) #& (njump > 0)
            hot &= (dark.data[integ,1,:,:] > hot_thresh[0]*2)
        else:
            med_rate = np.nanmedian(ds[:2,:,:])
            hot = (ds[0,:,:]-med_rate > hot_thresh[0])
            hot = (ds[1,:,:]-med_rate > hot_thresh[0])
        
        hmask = nd.binary_erosion(hot, iterations=4)
        hmask = nd.binary_dilation(hmask, iterations=4)
        hot &= ~hmask
        
        msg = f'process_uncal: {file} {integ+1} {hot_type} hot pixels > '
        msg += f' {hot_thresh[0]}: {hot.sum()}'
        utils.log_comment(utils.LOGFILE, msg, show_date=False, verbose=verbose)
        
        plus = np.array([[0,1,0],[1,1,1],[0,1,0]])
        hot_plus = nd.binary_dilation(hot, structure=np.ones((3,3)))

        hot = (ds[0,:,:] > hot_thresh[1]) & (njump > 0)
        
        hmask = nd.binary_erosion(hot, iterations=4)
        hmask = nd.binary_dilation(hmask, iterations=4)
        hot &= ~hmask
        
        msg = f'process_uncal: {file} {integ+1} very hot pixels > '
        msg += f' {hot_thresh[1]}: {hot.sum()}'
        utils.log_comment(utils.LOGFILE, msg, show_date=False, verbose=verbose)
        
        hot_plus |= nd.binary_dilation(hot, structure=np.ones((3,3)))
    
        hot_plus |= nd.binary_dilation((njump > 4)) #, structure=np.ones((3,3)))
        
        # Final slope image
        # sci[njump > 1] = np.nan
        dq |= hot_plus*DQ_HOT + (nbad > 1)*DQ_RESID
        sci[~np.isfinite(sci) | (nbad > 1) | (hot_plus)] = 0.
    
        # Uncertainty
        valid = (np.isfinite(sci) & (sci != 0))
        rmask = resid * valid
        rmask[rmask == 0] = np.nan
        
        # Effective integration time used for uncertainty estimates
        itime = dark.meta.exposure.group_time #*dark.meta.exposure.ngroups
        
        # Read noise derived from the ramp residuals
        if dark.meta.exposure.ngroups > groups_for_rnoise:
            rn_i = utils.nmad(rmask[np.isfinite(rmask) & valid])
            msg = f'process_uncal: {file} {integ+1} computed RNOISE: {rn_i:.2f}'
            utils.log_comment(utils.LOGFILE, msg, show_date=False, 
                              verbose=verbose)
            
            rn_i = np.ones_like(readnoise_2d)*rn_i
        else:
            rn_i = readnoise_2d
        
        rn_var = 12 * rn_i**2 / np.maximum(nsamp**3-nsamp, 6)
        p_var = np.maximum(slope*itime,0)/np.sqrt(nsamp)/gain_2d
        err = np.sqrt(rn_var + p_var)/itime
        
        dq[nsamp <= 2] |= DQ_ERR
        dq |= (~np.isfinite(err))*DQ_ERR
        dq |= (dq > 0)*DQ_ANY
    
        sci[dq > 0] = 0
        err[dq > 0] = 0
        
        slope_list.append(slope)
        slope_err_list.append(slope_err)
        
        sci_list.append(sci)
        err_list.append(err)
        dq_list.append(dq)
        nsamp_list.append(nsamp)
        rn_list.append(rn_i)
        jump_list.append(jump)
        
        # Set DQ=4 jump flag in groupdq
        for ij in range(NGROUPS-1):
            groupdq[integ,ij+1,:,:] |= (jump[ij,:,:]*4).astype(groupdq.dtype)
            
    ## Combine iterations
    for iter in range(3):
        #print(f'iter combine {iter}')
        sci = np.array(sci_list)
        err = np.array(err_list)
        dqi = np.array(dq_list)
        ivar = 1/err**2
        ivar[(dqi > 0) | (err == 0)] = 0
    
        num = (sci*ivar).sum(axis=0)
        den = ivar.sum(axis=0)
    
        sci = num/den
        err = 1/np.sqrt(den)

        for integ in range(NINTS):
            nsamp = nsamp_list[integ]
            rn_var = 12*rn_list[integ]**2/np.maximum(nsamp**3-nsamp, 6)
            _var = rn_var + np.maximum(sci*itime,0)/gain_2d/np.sqrt(nsamp)
            err_list[integ] = np.sqrt(_var)/itime
            err_list[integ][dq_list[integ] > 0] = 0
    
    dq = ((~np.isfinite(sci+err)) | (den <= 0))*1 | np.min(dqi, axis=0)
    dq *= 1
    
    if dark.meta.instrument.detector in PIXELDQ_BITS:
        pixeldq_bits = PIXELDQ_BITS[dark.meta.instrument.detector]
        
        msg = f'process_uncal: {file} set pixeldq bits {pixeldq_bits}'
        utils.log_comment(utils.LOGFILE, msg, show_date=False, verbose=verbose)
        dqp = np.zeros_like(dark.pixeldq)
        for b in pixeldq_bits:
            dqp |= dark.pixeldq & b
    
        dq |= dqp
    
    valid = dq == 0
    sci[~valid] = 0
    err[~valid] = 0
        
    ## Outlier rejection across integrations
    if NINTS >= outlier_min_nints:
        if rescale_uncertainty:
            sci_med = np.nanmedian(sci[valid])
            evalid = valid & (np.abs((sci - sci_med)/err) < 30)
            xerr_scale = utils.nmad(((sci - sci_med)/err)[evalid])
        else:
            xerr_scale = 1.
        
        msg = f'\nprocess_uncal: {file} find integration outliers '
        msg += f' sigmas={integration_sigmas}  err_scale={xerr_scale:.2f}'
        utils.log_comment(utils.LOGFILE, msg, show_date=False, verbose=verbose)
        
        for isig, sn_limit in enumerate(integration_sigmas):
            xsci = np.array(sci_list)
            xerr = np.array(err_list)
            xdqi = np.array(dq_list)
            
            resid = (xsci - sci)/xerr #/xerr_scale #xerr/xerr_scale
            xivar = 1/xerr**2
            if isig > 0:
                bad = np.abs(resid) > sn_limit
            else:
                bad = resid > sn_limit
            
            xivar[(dqi > 0) | (err == 0) | bad] = 0
    
            num = (xsci*xivar).sum(axis=0)
            den = xivar.sum(axis=0)
    
            sci = num/den
            err = 1/np.sqrt(den)
    
        dq = ((~np.isfinite(sci+err)) | (den <= 0))*1 | np.min(dqi, axis=0)
        valid = dq == 0
        sci[~valid] = 0
        err[~valid] = 0
    
    rn_meas = np.nanmedian(rn_list)
    
    nsamp = np.array(nsamp_list).sum(axis=0)
    rn_tot = np.nanmean(np.array(rn_list), axis=0)
    rn_var = 12*rn_tot**2/np.maximum(nsamp**3-nsamp, 6)/itime**2
    p_var = np.maximum(sci*itime,0)/gain_2d/itime**2
    
    if rescale_uncertainty:
        
        err = np.sqrt(rn_var + p_var)
        
        if rescale_with_background & HAS_SEP:
            bkg = sep.Background(sci.astype(np.float32), mask=(~valid), 
                                 **bkg_kwargs)
            sci_med = bkg.back()
        else:
            sci_med = np.nanmedian(sci[valid])
            
        evalid = valid & (np.abs((sci - sci_med)/err) < 30)
        evalid &= (err > 0) & np.isfinite(err)
        
        err_scale = utils.nmad(((sci - sci_med)/err)[evalid])
        msg = f'\nprocess_uncal: {file} uncertainty scale {err_scale:.2f}'
        utils.log_comment(utils.LOGFILE, msg, show_date=False, verbose=verbose)
        
        rn_var *= err_scale**2
        p_var *= err_scale**2
        err = np.sqrt(rn_var + p_var)
        
        #print('xxx', utils.nmad(((sci - sci_med)/err)[evalid]))
        #import matplotlib.pyplot as plt
        #plt.hist(((sci-sci_med)/err)[evalid],  bins=np.linspace(-10,10,128))
        
    else:
        err_scale = 1.
        evalid = valid
        
    with pyfits.open(file.replace('_uncal','_rate')) as im:
        imsci = im['SCI'].data*1
        imerr = im['ERR'].data*1

        im['SCI'].header['RNMEAS'] = (rn_meas,
                                      'Estimated read noise from ramps')
        # im['SCI'].header['RNDET'] = (rn_det, 'Assumed read noise')
        
        im['SCI'].data = sci.astype(im['SCI'].data.dtype)
        im['ERR'].data = err.astype(im['ERR'].data.dtype)
        im['ERR'].header['ESCALE'] = (err_scale, 'Uncertainty scale factor')
        im['ERR'].header['ESCLNPIX'] = (evalid.sum(),
                                        'Number of pixels used for scaling')
        
        im['DQ'].data = dq.astype(im['DQ'].data.dtype)
        im['VAR_RNOISE'].data = rn_var.astype(im['VAR_RNOISE'].data.dtype)
        im['VAR_POISSON'].data = p_var.astype(im['VAR_POISSON'].data.dtype)
        
        im.writeto(file.replace('_uncal', output_extension), overwrite=True)
        
        
    debug = {'file':file,
             'sci':sci, 'err':err, 'dq':dq,
             'nsamp':nsamp, 'rn':rn_meas,
             'times':times, 'dark':dark,
             'jump':jump,
             'jump_list':jump_list,
             'groupdq':groupdq,
             'njump':njump, 'resid':resid,
             'smodel':smodel, 'slope':slope, 'ahat':ahat,
             'slope_err':slope_err,
             'raw_err':raw_err,
             'hot_plus':hot_plus,
             'slope_list':slope_list,
             'slope_err_list':slope_err_list,
             'sci_list':sci_list,
             'err_list':err_list,
             'dq_list':dq_list,
             'jump_saturated':jump_saturated,
             'gain_2d':gain_2d,
             'readnoise_2d':readnoise_2d,
             'imsci':imsci,'imerr':imerr,
        }
    
    # Reset LOGFILE
    utils.LOGFILE = _LOGFILE
    warnings.filterwarnings('default', category=RuntimeWarning)
    
    return debug


def grow_snowball_jumps(dark, jump, erode_snowballs=3, grow_snowballs=5, verbose=True, **kwargs):
    """
    Grow masks around large groups of pixels identified in the jump step that
    are likely "snowballs"
    
    Parameters
    ----------
    dark : 
    
    jump : 
    
    erode_snowballs : int
        Number of `scipy.ndimage.binary_erosion` iterations run on the jump
        mask of each read before identifying snowballs.  This removes small
        clumps of pixels that probably aren't big snowballs
    
    grow_snowballs : float
        Once snowballs are identified as big clumps of jump pixels, compute
        the pixel area of each smowball clump and make a circular mask with
        radius ``grow_snowballs * np.sqrt(area/pi)``
    
    Returns
    -------
    group_dq :array-like
        DQ mask array with dimensions ``NINT, NGROUP, NY, NX`` with the ``4``
        bit set for identified jumps
    
    """
    nint, ngroup, NYPIX, NXPIX = jump.groupdq.shape
    
    groupdq = (jump.groupdq*1).astype(jump.groupdq.dtype)
    
    jump_saturated = np.zeros((NYPIX, NXPIX), dtype=bool)
    
    for ii in range(nint):
        for ig in range(1,ngroup):
            ji = groupdq[ii,ig,:,:] & 4 > 0
            sat = dark.groupdq[ii,ig:,:,:].max(axis=0) > 0
            ji |= sat
            
            js = nd.binary_erosion(ji, iterations=erode_snowballs)
            
            jump_saturated |= js # | sat
            
            groupdq[ii,ig,:,:] |= (jump_saturated*2).astype(groupdq.dtype)
            
            label, num_labels = nd.label(js)
            if num_labels > 0:
                
                msg = f'process_uncal: {ii+1} snowball read {ig:>2} '
                msg += f' n={num_labels:>2}  '
                msg += f'nsat={int(jump_saturated.sum())//4:>6}'
                utils.log_comment(utils.LOGFILE, msg, show_date=False, 
                                  verbose=verbose)
                
                xy_label = nd.center_of_mass(label, labels=label, 
                                             index=np.unique(label)[1:])
                area = nd.sum(label > 0, labels=label, 
                              index=np.unique(label)[1:])

                R_label = np.sqrt(area/np.pi)
                yp, xp = np.indices((NYPIX, NXPIX))

                blob = np.zeros_like(js)
                #smask = np.zeros_like(js)

                for (yyi, xxi), ri in zip(xy_label, R_label):
                    R = np.sqrt((xp-xxi)**2 + (yp-yyi)**2)
                    blob |= R < grow_snowballs*np.maximum(ri, 2)
                    #smask |= R < grow_snowballs/2*np.maximum(ri, 2)
                
                groupdq[ii,ig,:,:] |= (blob*4).astype(groupdq.dtype)
    
    return groupdq


def get_level1_references(model):
    """
    Get CRDS metadata for reference files
    
    Parameters
    ----------
    model : `jwst.datamodels.ImageModel`
        Datamodel created by the pipeline steps, e.g., ``gain_step``
    
    Returns
    -------
    ref_file : list
        List of tuples of ``(step, reference_filename)``
    
    """    
    pipe = get_pipeline_object()
    
    steps = [pipe.group_scale,
             pipe.dq_init ,
             pipe.saturation,
             pipe.superbias,
             pipe.refpix,
             pipe.linearity,
             pipe.persistence,
             pipe.dark_current,
             pipe.jump,
             pipe.ramp_fit,
             pipe.gain_scale]
    
    ref_files = []
    keys = []
    for step in steps:
        for reftype in step.reference_file_types:
            _file = pipe.get_reference_file(model, reftype)
            #print(reftype, files)
            if _file == 'N/A':
                continue
            else:
                row = (reftype, 'crds://' + os.path.basename(_file))
                if reftype not in keys:
                    ref_files.append(row)
                    keys.append(reftype)
    
    return ref_files


def pipeline_level1(file, output_extension='_rate', maximum_cores='half', use_pipe_jumps=True, unset_jump_dq=True, force_uncal=True, debug=None, **kwargs):
    """
    Use the pipeline for the ramp fits with manual jump detections
    
    Parameters
    ----------
    file : str
        JWST raw "uncal" filename
    
    output_extension : str
        The script puts the derived ramp and uncertainty into a ``rate``-like 
        file with filename ``file.replace('_uncal', output_extension)``.
    
    maximum_cores : 'none', 'quarter', 'half', 'all'
        Number of processes for `jwst` jump and ramp pipeline steps
    
    unset_jump_dq : bool
        Unset the ``DQ=4`` pixels from the final product
    
    force_uncal : bool
        Require that ``file`` endswith "_uncal.fits"
    
    debug : dict, bool
        Previous output
    
    kwargs : dict
        Keyword arguments passed to `grizli.jwst_level1.process_uncal_level1`
    
    Returns
    -------
    debug : dict
        Full output from `grizli.jwst_level1.process_uncal_level1`
    
    gain : `jwst.datamodels.ImageModel`
        Result of the final `gain_scale` level 1 step, after the others
        have been run (dark, jump, etc.).
    
    """    
    #from jwst.pipeline import Detector1Pipeline
    frame = inspect.currentframe()
    _LOGFILE = utils.LOGFILE
    warnings.filterwarnings('ignore', category=RuntimeWarning)

    utils.LOGFILE = file.split('.fits')[0] + '.log'
    utils.log_function_arguments(utils.LOGFILE, frame, 'process_uncal',
                                 ignore=['debug','_LOGFILE'],
                                 )
    
    if (not file.endswith('_uncal.fits')) & (force_uncal):
        utils.LOGFILE = _LOGFILE
        warnings.filterwarnings('default', category=RuntimeWarning)
        
        raise ValueError(f'{file} does not have _uncal extension')
        
        
    #pipe = Detector1Pipeline()
    pipe = get_pipeline_object()
    
    dark = process_level1_to_dark(file)
    pipe.jump.maximum_cores = maximum_cores
    jump = pipe.jump.process(dark)
    
    jumpdq = grow_snowball_jumps(dark, jump, **kwargs)
    jump.groupdq |= jumpdq
    
    # Unset some DQ flags    
    orig_dq = jump.pixeldq*1
    
    dq = np.zeros_like(jump.pixeldq)
    for i in range(1,20):
        #for i in PIXELDQ_BITS[dark.meta.instrument.detector]:
        if i in [15,]:
            continue
            
        dqi = 2**i & jump.pixeldq
        dq |= dqi
        
    jump.pixeldq = dq
    
    # Run Ramp_fit
    pipe.ramp_fit.maximum_cores = maximum_cores
    myramp = pipe.ramp_fit.process(jump)
    
    # Run Gain
    gain = pipe.gain_scale.process(myramp[0])
    
    bad = ~np.isfinite(gain.data + gain.err)
    bad |= dq > 0
    #bad |= (gain.dq & 1) == 1
    gain.dq[bad] |= 1
    gain.data[bad] = 0.
    gain.err[bad] = 0
    
    if unset_jump_dq:
        gain.dq -= (gain.dq & 4)
    
    # Update metadata with reference files since this doesn't seem to
    # be run automatically?
    ref_files = get_level1_references(gain)
    pipe.gain_scale.finalize_result(gain, ref_files)
    
    gain.write(file.replace('_uncal', output_extension),
               overwrite=True)
    
    
    utils.LOGFILE = _LOGFILE
    warnings.filterwarnings('default', category=RuntimeWarning)
    
    return debug, gain